using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.Reflection;
using Produmex.Foundation;
using Produmex.Foundation.Collections;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.ProLaser.BusinessObjects.Definitions.Tables;

// PmxWorkflowExecutionTypes.HOOK_FLOW
namespace WorkflowScript_AfterPickListPackedHookScript
{
    /// <summary>
    /// This is a hook script, meant to be customized on a "per project" basis
    /// </summary>
    public class WorkflowScript_AfterPickListPackedHookScript : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
        public ReadOnlyBinder<int> PickListDocEntry;
        public ReadWriteBinder<bool> DoAutomaticShipping;

        // Output parameters * do not change *
        public ReadWriteBinder<PmxDictionary<string, object>> DeliveryData;

        // Sub-flows
        WorkflowInstanceBase m_printDeliveryDocumentsSubFlow;
        protected int? m_subFlowPickListEntry;

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_AfterPickListPackedHookScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_AfterPickListPackedHookScript( WorkflowInstanceBase parent, WorkflowInstanceFactory factory )
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
			// Parameters in scope
			Session session = GetScopeParameter("Session") as Session;
			ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            if (DoAutomaticShipping.Get())
            {
                //Always set to false!
                DoAutomaticShipping.Set(false);

                //We need to do automatic shipping, so we need to check if this pick list is the last pick list for the current cardcode, shipto address, and delivery time
                string storageLocation = null;
                List<int> listOfPickLists = GetAllPackedPickLists(sboProviderService, out storageLocation);
                if (listOfPickLists.Count > 0)
                {
                    DocumentRef[] deliveryDocRefs = MakeDeliveries(listOfPickLists, sboProviderService, session);

                    if (deliveryDocRefs != null &&
                        deliveryDocRefs.Length > 0)
                    {
                        PrintDeliveryDocuments(deliveryDocRefs, listOfPickLists, storageLocation, sboProviderService, session);
                    }
                }
            }


        }

        private void PrintDeliveryDocuments(DocumentRef[] deliveryDocRefs, List<int> listOfPickLists, string storageLocation, ISboProviderService sboProviderService, Session session)
        {
            try
            {
                m_subFlowPickListEntry = PickListDocEntry.Get();

                m_printDeliveryDocumentsSubFlow = this.Create("PrintSalesDeliveryDocumentsScript");
                m_printDeliveryDocumentsSubFlow.BindParameters(
                    this.DefaultCultureInfo,
                    this.PmxOseCompany,
                    new ReadOnlyBinder<DocumentRef[]>("ListOfDeliveries", deliveryDocRefs),
                    new ReadWriteBinder<int?>("PickListEntry", this, "m_subFlowPickListEntry"),
                    new ReadOnlyBinder<string>("LocationCode", storageLocation));
                this.StartSubFlowAndWait(m_printDeliveryDocumentsSubFlow, true);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }
            }

        }

        private DocumentRef[] MakeDeliveries(List<int> listOfPickLists, ISboProviderService sboProviderService, Session session)
        {
            ProdumexError pmxError = null;
            Produmex.Foundation.Messages.Message msg = null;
            DocumentRef[] deliveryDocRefs = null;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
              this.DefaultCultureInfo, BuildParamCollection(
              "MessageKey", "MSG_ADDING_SALES_DELIVERIES"));

            while (true)
            {
                try
                {
                    //Make automatic sales delivery
                    deliveryDocRefs = sboProviderService.InvokeMethodWithDbConnection<DocumentRef[]>(true, false, null, null,
                        delegate (PmxDbConnection conn, object[] parameters)
                        {
                            PmxSalesDeliveryNoteProvider sdnProv = new PmxSalesDeliveryNoteProvider(conn);
                            return sdnProv.ShipAllItemsForPickList(new Collection<int>(listOfPickLists), null, null);
                        });


                    break;
                }
                catch (Exception ex)
                {
                    s_log.Error("", ex);
                    if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, DefaultCultureInfo.Get()),
                                "NoTranslationOfMessageKey", true,
                                "IsError", true,
                                "ShowButton", true));
                        WaitForMessage();


                    }
                    else
                    {
                        // Inform the user with the error
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            DefaultCultureInfo.Get(), BuildParamCollection(
                                "PmxError", ((ProdumexException)ex).Error,
                                "ShowButton", true));
                        WaitForMessage();

                    }

                    //ask user if he wants to try again
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "TitleKey", "MSG_TITLE_RETRY_UPDATE",
                            "MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
                    msg = WaitForMessage();

                    if (!msg.Name.ToLower().EndsWith(".yes"))
                    {
                        break;
                    }
                }

            }

            return deliveryDocRefs;
        }

        private List<int> GetAllPackedPickLists(ISboProviderService sboProviderService, out string locationCode)
        {
            locationCode = null;
            List<int> listOfPickLists = new List<int>();
            string query = BuildQuery.GetAllPickListsForSameCustomer(PickListDocEntry.Get(), sboProviderService.GetDbTool());

            List<string> listOfAllowedPickListStatus = new List<string>() { PmxPickListStatusColumnValues.Packed.VALUE.ToString(), PmxPickListStatusColumnValues.PartiallyDelivered.VALUE.ToString() };

            DataSet ds = sboProviderService.RunView(false, null, null, query);

            if (GetNumberOfRows(ds) > 0)
            {
                foreach (DataRow row in ds.Tables[0].Rows)
                {
                    if (string.IsNullOrEmpty(locationCode))
                    {
                        locationCode = row[PmxPickListTable.Columns.DestStorLocCode.NAME] as string;
                    }

                    if (listOfAllowedPickListStatus.Contains(row[PmxPickListTable.Columns.PickListStatus.NAME] as string))
                    {
                        //Pick list is packed, so add to list of pick lists that need to be automatically shipped
                        listOfPickLists.Add(Convert.ToInt32(row[PmxPickListTable.Columns.DocEntry.NAME]));
                    }
                    else
                    {
                        //Cannot do automatic shipping, because not all pick lists are packed
                        listOfPickLists.Clear();
                        break;
                    }
                }
            }



            return listOfPickLists;
        }

        #endregion


        private static int GetNumberOfRows(DataSet ds)
		{
			if (ds != null && ds.Tables.Count > 0)
			{
				return ds.Tables[0].Rows.Count;
			}

			return 0;
		}

		#region Query

		private static class BuildQuery
		{
			public static string GetAllPickListsForSameCustomer(int pickListDocEntry, DbTool dbTool)
			{
				DbQueryBuilder qb = dbTool.GetQueryBuilder();




			    string linkTableName = "LINK_" + PmxPickListTable.NAME;

                qb = qb.Select.Distinct
						.EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME).Comma
                        .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.PickListStatus.NAME).Comma
                        .EscTableColumnName(linkTableName, PmxPickListTable.Columns.DestStorLocCode.NAME).CRLF
                    .From
						.EscTableName(PmxPickListTable.NAME).WithNoLock.CRLF
                    .InnerJoin
                        .EscTableName(PmxPickListProposalLineTable.NAME).WithNoLock
                        .On.EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.PickListProposalEntry.NAME).Eq
                        .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME).CRLF
                    .InnerJoin
                        .EscTableName(PmxSalesOrderLineTable.NAME).WithNoLock
                        .On.EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.DocEntry.NAME).Eq
                        .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME).CRLF
                        .And.EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME).Eq.String_(SboObjectTypes.SalesOrder).CRLF
                        .And.EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.LineNum.NAME).Eq
                        .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseLine.NAME).CRLF
                    .InnerJoin
                        .EscTableName(PmxShippingTypeTablePL.NAME).WithNoLock
                        .On.EscTableColumnName(PmxShippingTypeTablePL.NAME, PmxShippingTypeTablePL.Columns.Code.NAME).Eq
                        .EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.TrnsCode.NAME).CRLF
                    .InnerJoin.Append("(")
                            .Select
			                .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME).Comma
                            .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.ShipToAddress.NAME).Comma
                            .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.DestStorLocCode.NAME).Comma
                            .EscTableColumnName(PmxShippingTypeTablePL.NAME, PmxShippingTypeTablePL.Columns.DeliveryTime.NAME).Comma
			                .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.CardCode.NAME).CRLF
			                .From
			                .EscTableName(PmxPickListTable.NAME).WithNoLock.CRLF
			                .InnerJoin
			                    .EscTableName(PmxPickListProposalLineTable.NAME).WithNoLock
			                    .On.EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.PickListProposalEntry.NAME).Eq
			                    .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME).CRLF
			                .InnerJoin
			                    .EscTableName(PmxSalesOrderLineTable.NAME).WithNoLock
			                    .On.EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.DocEntry.NAME).Eq
			                    .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME).CRLF
			                    .And.EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME).Eq.String_(SboObjectTypes.SalesOrder).CRLF
			                    .And.EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.LineNum.NAME).Eq
			                    .EscTableColumnName(PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseLine.NAME).CRLF
			                .InnerJoin
			                    .EscTableName(PmxShippingTypeTablePL.NAME).WithNoLock
			                    .On.EscTableColumnName(PmxShippingTypeTablePL.NAME, PmxShippingTypeTablePL.Columns.Code.NAME).Eq
			                    .EscTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.TrnsCode.NAME).CRLF
                    .Append(")").As.EscTableName(linkTableName)
                        .On.EscTableColumnName(linkTableName, PmxPickListTable.Columns.CardCode.NAME).Eq
                        .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.CardCode.NAME).CRLF
                        .And.EscTableColumnName(linkTableName, PmxPickListTable.Columns.DocEntry.NAME).Eq.Append(qb.AsOwnType(pickListDocEntry)).CRLF
                        .And.EscTableColumnName(linkTableName, PmxPickListTable.Columns.ShipToAddress.NAME).Eq
                        .EscTableColumnName(PmxPickListTable.NAME, PmxPickListTable.Columns.ShipToAddress.NAME).CRLF
                        .And.Append($"{qb.IFNULL}(").EscTableColumnName(linkTableName, PmxShippingTypeTablePL.Columns.DeliveryTime.NAME).Append(", '')").Eq
                        .Append($"{qb.IFNULL}(").EscTableColumnName(PmxShippingTypeTablePL.NAME, PmxShippingTypeTablePL.Columns.DeliveryTime.NAME).Append(", '')").CRLF
                    .Where
						.EscColumnName(PmxPickListTable.Columns.DocStatus.NAME)
						.Eq
						.String_(SboDocStatusColumnValues.Open.VALUE);

				return qb.ToString();
			}
		}

		#endregion
	}
}
